use super::*;
use std::sync::Arc;

use data::Cache;
use config::Config;

/// Program state
#[derive(Debug, Clone)]
pub struct State
{
    config: Arc<Config>,
    cache: Cache,
    depth: usize,
}

impl State
{
    /// Create a new state
    pub fn new(cfg: Config) -> Self
    {
	Self {
	    config: Arc::new(cfg),
	    cache: Cache::new(),
	    depth: 1,
	}
    }

    /// Current depth of this tree
    pub fn depth(&self) -> usize
    {
	self.depth
    }
    
    /// Clone into increased depth, if config allows a deeper run.
    pub fn deeper(&self) -> Option<Self>
    {
	if self.config.recursive.can_run(self.depth+1) {
	    Some(Self{
		depth: self.depth + 1,
		..self.clone()
	    })
	} else {
	    None
	}
    }

    /// The configuration for this run
    pub fn config(&self) -> &Config
    {
	&self.config
    }

    /// A reference to the cache of this run
    pub fn cache(&self) -> &Cache
    {
	&self.cache
    }

    /// Subscribe to this state's cache
    pub fn cache_sub(&self) -> Cache
    {
	self.cache.clone()
    }

    /// Try to consume the state into the cache.
    ///
    /// Fails if there are other references of this state alive.
    pub fn try_into_cache(self) -> Result<Cache, Self>
    {
	match Arc::try_unwrap(self.config)
	{
	    Ok(_) => Ok(self.cache),
	    Err(config) => Err(Self{config, ..self}),
	}
    }
}
